function [vartheta,u] = TOALocByRB_CFS_KnownSo(r,Tr,C,so,Q_r,Q_so)
% [vartheta,u] = TOALocByRB_CFS_KnownSo(r,Tr,C,so,Q_r,Q_so)
%
% Closed-form solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when a noisy RB position is available.
%
% Input:
%   r:        (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%              M = number of sensors, N = number of transmitters
%   Tr:       (3 x N), transmitter positions
%   C:        (3 x M), sensor positions local to the RB
%   so:       (3 x 1), RB position measurment
%   Q_r:      (NM x 1), covariance matrix of distance measurements
%   Q_so:     (3 x 1), covariance matrix of RB position measurment
%
% Output:
%   vartheta: (8 x 1), estimate of (global) object position, direction angles of local object position, and RB position
%   u:        (3 x 1), object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                  % localization dimension
Q = blkdiag(Q_r,Q_so);  % covariance matrix of composite observation vector 
[~,N] = size(Tr);       % N = number of transmitters
[~,M] = size(C);        % M = number of sensors

% -- first stage --
Mr = zeros(M*N,N);
for k = 1:1:N
    Mr((k-1)*M+1:k*M,k) = r((k-1)*M+1:k*M);
end
h1 = r.^2 + kron( (Tr.^2)'*ones(K,1), ones(M,1) ) - kron( ones(N,1), (C.^2)'*ones(K,1) ) - norm(so)^2;
G1 = 2*[kron(Tr'-ones(N,1)*so',ones(M,1)), -kron(ones(N,1),C'),zeros(M*N,3),Mr;
        zeros(3,6),0.5*eye(3),zeros(3,N)];
h1 = [h1;so];
y = (G1'*G1)\G1'*h1; 
for ii = 1:1:2
    d = y(10:9+N);
    B1  = [2*diag(r-kron(d,ones(M,1))),2*ones(M*N,1)*(y(1:K)-so)';
           zeros(3,N*M),eye(K)];
    Q1 = B1*Q*B1';
    y = (G1'/Q1*G1)\G1'/Q1*h1; % first stage solution
end

% -- second stage --
h2 = [y(1:3); norm(y(4:6))^2;atan2(y(5),y(4)); asin(y(6)/norm(y(4:6)));
      y(7:9); y(10:N+9).^2-(Tr.^2)'*ones(K,1)];
G2 = [eye(K),zeros(K,2+K);
      y(1:3)'-y(7:9)',zeros(1,2),-y(1:3)'+y(7:9)';
      zeros(2,3),eye(2),zeros(2,K);
      zeros(K,5),eye(K);
      ones(N,1)*y(1:K)'-2*Tr',zeros(N,2+K)];
p = y(4:6);
e3 = [0,0,1];
Jphi_p = [-p(2)/(norm(p(1:2))^2),p(1)/(norm(p(1:2))^2),0;
     -p(3)*p'/(norm(p)^2*norm(p(1:2)))+e3/norm(p(1:2))];
B2 = [eye(K),zeros(K,N+2*K);
     -y(1:3)'+y(7:9)',2*y(4:6)',y(1:3)'-y(7:9)',zeros(1,N);
     zeros(2,3),Jphi_p,zeros(2,3+N);
     zeros(3,6),eye(3),zeros(3,N);
     -ones(N,1)*y(1:K)',zeros(N,2*K),2*diag(d)];
W2 = (B2')\(G1'/Q1*G1)/B2;
vartheta = (G2'*W2*G2)\G2'*W2*h2;
u = vartheta(1:K);

end