function [theta,u] = TOALocByRB_CFS_UnknownSo(r,Tr,C,Q_r)
% [theta,u] = TOALocByRB_CFS_UnknownSo(r,Tr,C,Q_r)
%
% Closed-form solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when the RB position is fully unknown.
%
% Input:
%   r:        (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%              M = number of sensors, N = number of transmitters
%   Tr:       (3 x N), transmitter positions
%   C:        (3 x M), sensor positions local to the RB
%   Q_r:      (NM x 1), covariance matrix of distance measurements
%
% Output:
%   theta:    (6 x 1), estimate of (global) object position and local object position.
%   u:        (3 x 1), object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                  % localization dimension
[~,N] = size(Tr);       % N = number of transmitters
[~,M] = size(C);        % M = number of sensors

% -- first stage --
Mr = zeros(M*N,N);
for k = 1:1:N
    Mr((k-1)*M+1:k*M,k) = r((k-1)*M+1:k*M);
end
f1 = r.^2 + kron( (Tr.^2)'*ones(K,1), ones(M,1) ) - kron( ones(N,1), (C.^2)'*ones(K,1) );
H1 = 2*[kron(Tr',ones(M,1)),Mr, -kron(ones(N,1),C'),0.5*ones(M*N,1)];
x = (H1'*H1)\H1'*f1;
for ii = 1:1:2
    d = x(K+1:K+N);
    D1  = 2*diag(r-kron(d,ones(M,1)));
    Q1 = D1*Q_r*D1';
    x = (H1'/Q1*H1)\H1'/Q1*f1;
end

% -- second stage --
f2 = [x(1:K);x(K+1:N+K).^2-(Tr.^2)'*ones(K,1);x(K+N+1:N+2*K);norm(x(K+N+1:N+2*K))^2-x(N+2*K+1)];
H2 = [eye(K),zeros(K,K);
      ones(N,1)*x(1:K)'-2*Tr',zeros(N,K);
      zeros(K,3),eye(K);
      x(1:K)',zeros(1,K)];
D2 = [eye(K),zeros(K,N+K+1);
      -ones(N,1)*x(1:K)',2*diag(d),zeros(N,1+K);
      zeros(K,K+N),eye(K),zeros(K,1);
      -x(1:K)',zeros(1,N),2*x(K+N+1:N+2*K)',-1];
W2 = (D2')\(H1'/Q1*H1)/D2;
theta = (H2'*W2*H2)\H2'*W2*f2;
u = theta(1:K);

end


