function u = TOALocByRB_MLE_KnownSo(r,Tr,C,so,Q_r,Q_so,MaxIter,vartheta0)
% u = TOALocByRB_MLE_KnownSo(r,Tr,C,so,Q_r,Q_so,MaxIter,vartheta0)
%
% Gauss-Newton MLE solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when a noisy RB position is available.
%
% Input:
%   r:           (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%                 M = number of sensors, N = number of transmitters
%   Tr:          (3 x N), transmitter positions
%   C:           (3 x M), sensor positions local to the RB
%   so:          (3 x 1), RB position measurment
%   Q_r:         (NM x 1), covariance matrix of distance measurements
%   Q_so:        (3 x 1), covariance matrix of RB position measurment
%   MaxIter:     (1 x 1), maximum number of iterations
%   vartheta0:   (8 x 1), initial solution of object position, direction angles of local object position, and RB position
%
% Output:
%   u:           (3 x 1), object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                  % localization dimension
[~,N] = size(Tr);       % N = number of transmitters
[~,M] = size(C);        % M = number of sensors
Q = blkdiag(Q_r,Q_so);  % covariance matrix of composite observation vector 
Q = (N*M+K)*Q/trace(Q); % scaling the covariance matrix

vartheta = vartheta0;
for iter = 1:1:MaxIter
    u = vartheta(1:K);
    rho_p = [cos(vartheta(K+1))*cos(vartheta(K+2));sin(vartheta(K+1))*cos(vartheta(K+2));sin(vartheta(K+2))];
    p = norm(vartheta(1:K)-vartheta(2*K:end))*rho_p;
    hat_so = vartheta(end-K+1:end);
    hat_r = zeros(M,1);
    Jr_u = zeros(M,K);
    Jr_phi = zeros(M,K-1);
    Jr_so = zeros(M,K);
    for i=1:1:N
        for j = 1:1:M
            hat_r((i-1)*M+j) = norm(u - Tr(:,i))+norm(p - C(:,j));
            Jr_u((i-1)*M+j,:) = (u - Tr(:,i))'/norm(u - Tr(:,i))+(p - C(:,j))'*p*(u-hat_so)'/(norm(p)^2*norm(p - C(:,j)));
            Jr_phi((i-1)*M+j,1) = norm(p)*(p - C(:,j))'*[-sin(vartheta(K+1))*cos(vartheta(K+2));cos(vartheta(K+1))*cos(vartheta(K+2));0]/(norm(p - C(:,j)));
            Jr_phi((i-1)*M+j,2) = norm(p)*(p - C(:,j))'*[-cos(vartheta(K+1))*sin(vartheta(K+2));-sin(vartheta(K+1))*sin(vartheta(K+2));cos(vartheta(K+2))]/(norm(p - C(:,j)));
            Jr_so((i-1)*M+j,:) = -(p - C(:,j))'*p*(u-hat_so)'/(norm(p)^2*norm(p - C(:,j)));
        end
    end
    Jm_vartheta = [Jr_u,Jr_phi,Jr_so;zeros(K,2*K-1),eye(K)];
    delta_vartheta = inv(Jm_vartheta'*inv(Q)*Jm_vartheta)*Jm_vartheta'*inv(Q)*([hat_r;hat_so]-[r;so]);
    if (norm(delta_vartheta)>1e4)||(norm(delta_vartheta)<1e-5)
        break;
    else
        vartheta = vartheta - delta_vartheta;
    end
end

end

