function u = TOALocByRB_MLE_UnknownSo(r,Tr,C,Q_r,MaxIter,theta0)
% u = TOALocByRB_MLE_UnknownSo(r,Tr,C,Q_r,MaxIter,theta0)
%
% Gauss-Newton MLE solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when the RB position is fully unknown.
%
% Input:
%   r:           (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%                 M = number of sensors, N = number of transmitters
%   Tr:          (3 x N), transmitter positions
%   C:           (3 x M), sensor positions local to the RB
%   Q_r:         (NM x 1), covariance matrix of distance measurements
%   MaxIter:     (1 x 1), maximum number of iterations
%   theta0:      (6 x 1), initial solution of (global) object position and local object position
%
% Output:
%   u:           (3 x 1) object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                   % localization dimension
[~,N] = size(Tr);        % N = number of transmitters
[~,M] = size(C);         % M = number of sensors
Q_r = (N*M)*Q_r/trace(Q_r); % scaling the covariance matrix

theta = theta0;
for iter = 1:1:MaxIter
    hat_r = zeros(M*N,1);
    Jdt_u = zeros(M*N,K);
    Jds_p = zeros(M*N,K);
    u = theta(1:K);
    p = theta(K+1:2*K);
    for i=1:1:N
        for j = 1:1:M
            hat_r((i-1)*M+j) = norm(u - Tr(:,i))+norm(p - C(:,j));
            Jdt_u((i-1)*M+j,:) = (u - Tr(:,i))'/norm(u - Tr(:,i));
            Jds_p((i-1)*M+j,:) = (p - C(:,j))'/norm(p - C(:,j));
        end
    end
    Jr_theta = [Jdt_u,Jds_p];
    delta_theta =  inv(Jr_theta'*inv(Q_r)*Jr_theta)*Jr_theta'*inv(Q_r)*(hat_r-r);
    if (norm(delta_theta)>1e4)||(norm(delta_theta)<1e-5)
        break;
    else
        theta = theta-delta_theta;
    end
end

end

