function u = TOALocByRB_SDP_KnownSo(r,Tr,C,so,Q_r,Q_so)
% u = TOALocByRB_SDP_KnownSo(r,Tr,C,so,Q_r,Q_so)
%
% SDP solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when a noisy RB position is available.
%
% Input:
%   r:        (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%              M = number of sensors, N = number of transmitters
%   Tr:       (3 x N), transmitter positions
%   C:        (3 x M), sensor positions local to the RB
%   so:       (3 x 1), RB position measurment
%   Q_r:      (NM x 1), covariance matrix of distance measurements
%   Q_so:     (3 x 1), covariance matrix of RB position measurment
%
% Output:
%   u:        (3 x 1), object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                  % localization dimension
Q = blkdiag(Q_r,Q_so);  % covariance matrix of composite observation vector 
[~,N] = size(Tr);       % N = number of transmitters
[~,M] = size(C);        % M = number of sensors

% -- scaling the dataset --
scale = 1000;  % scaling factor for SDP
Tr = Tr/scale;
C = C/scale;
so = so/scale;
r = r/scale;

Mr = zeros(M*N,N);
for k = 1:1:N
    Mr((k-1)*M+1:k*M,k) = r((k-1)*M+1:k*M);
end
h1 = [r.^2 + kron( (Tr.^2)'*ones(K,1), ones(M,1) ) - kron( ones(N,1), (C.^2)'*ones(K,1) ) - norm(so)^2;
      so];
G1 = 2*[kron(Tr'-ones(N,1)*so',ones(M,1)),Mr, -kron(ones(N,1),C'),zeros(M*N,3);
        zeros(3,N+2*K),0.5*eye(K)];
Q1 = eye(M*N+K);
for i =1:1:2
    L = 3*K+N;
    F = [G1'/Q1*G1, -G1'/Q1*h1; -h1'/Q1*G1,1];
    
    % -- solve SDP by cvx --
    cvx_clear
    cvx_solver sedumi
    cvx_precision best
    cvx_begin sdp quiet;
    variable Y(L,L) symmetric;
    variable y(L);
    minimize( trace(F*[Y,y;y',1]) )
    subject to
    [Y y;y' 1]==semidefinite(L+1);
    trace(Y(K+N+1:2*K+N,K+N+1:2*K+N)) == trace( Y(1:K,1:K)-2*Y(1:K,2*K+N+1:3*K+N)+Y(2*K+N+1:3*K+N,2*K+N+1:3*K+N) );
    for i =1:N
        Y(K+i,K+i)==trace(Y(1:K,1:K)) - 2*Tr(:,i)'*y(1:K)+norm(Tr(:,i))^2;
        y(K+i)>=0;
        for j = i+1:1:N
            Y(K+i,K+j)>=abs( trace(Y(1:K,1:K)) - (Tr(:,i)+Tr(:,j))'*y(1:K)+Tr(:,i)'*Tr(:,j) );
        end
    end
    cvx_end
    
    % -- update the weighting matrix --
    d = y(K+1:K+N);
    B1  = [2*diag(r-kron(d,ones(M,1))),2*ones(M*N,1)*(y(1:K)-y(2*K+N+1:end))';
        zeros(3,N*M),eye(K)];
    Q1 = B1*Q*B1';
    
end
u = y(1:K)*scale;
end