function u = TOALocByRB_SDP_UnknownSo(r,Tr,C,Q_r)
% u = TOALocByRB_SDP_UnknownSo(r,Tr,C,Q_r)
%
% SDP solution for object localization in 3-D by propagation time 
% delays (TDs) to a rigid body receiver when the RB position is fully unknown.
%
% Input:
%   r:        (NM x 1), distance measurements (TD measurements times signal propagation speed), 
%              M = number of sensors, N = number of transmitters
%   Tr:       (3 x N), transmitter positions
%   C:        (3 x M), sensor positions local to the RB
%   Q_r:      (NM x 1), covariance matrix of distance measurements
%
% Output:
%   u:        (3 x 1), object position estimate
%
% Reference: X. Ke and K. C. Ho, "Localization of a passive object in 3-D by 
% propagation time delays to A rigid body receiver of unknown location," 
% IEEE Trans. Signal Process., vol. 72, pp. 3470-3486, 2024.
%
% X. Ke, K. C. Ho     8-28-2024
%
%       Copyright (C) 2024
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA.
%       hod@missouri.edu

K = 3;                  % localization dimension
[~,N] = size(Tr);       % N = number of transmitters
[~,M] = size(C);        % M = number of sensors

% -- scaling the dataset --
scale = 100;  % scaling factor for SDP
Tr = Tr/scale;
C = C/scale;
r = r/scale;

Mr = zeros(M*N,N);
for k = 1:1:N
    Mr((k-1)*M+1:k*M,k) = r((k-1)*M+1:k*M);
end
f1 = r.^2 + kron( (Tr.^2)'*ones(K,1), ones(M,1) ) - kron( ones(N,1), (C.^2)'*ones(K,1) );
H1 = 2*[kron(Tr',ones(M,1)),-kron(ones(N,1),C'),Mr,0.5*ones(M*N,1)];
Q1 = eye(M*N);
for i =1:1:2
    L = 7+N;
    F = [H1'/Q1*H1, -H1'/Q1*f1; -f1'/Q1*H1,1];
    
    % -- solve SDP by cvx --
    cvx_clear
    cvx_solver sedumi
    cvx_precision best
    cvx_begin sdp quiet;
    variable Y(L,L) symmetric;
    variable y(L);
    minimize( trace(F*[Y,y;y',1]) )
    subject to
    [Y y;y' 1]==semidefinite(L+1);
    y(N+7) == trace( Y(4:6,4:6) ) - trace( Y(1:3,1:3) );
    for i =1:N
        Y(6+i,6+i)==trace(Y(1:K,1:K)) - 2*Tr(:,i)'*y(1:K)+norm(Tr(:,i))^2;
        y(6+i)>=0;
        for j = i+1:1:N
            Y(6+i,6+j)>=abs( trace(Y(1:K,1:K)) - (Tr(:,i)+Tr(:,j))'*y(1:K)+Tr(:,i)'*Tr(:,j) );
        end
    end
    cvx_end
    
    % -- update the weighting matrix --
    d = y(7:6+N);
    B1  = 2*diag(r-kron(d,ones(M,1)));
    Q1 = B1*Q_r*B1';
    
end
u = y(1:K)*scale;
end